(**************************************************************************)
(*                                                                        *)
(*                               GlSurf                                   *)
(*                                                                        *)
(*                   C. Raffalli, Universite de Savoie                    *)
(*                                                                        *)
(* Copyright 2003, 2004 Christohe Raffalli                                *)
(*                                                                        *)
(*  This file is part of GlSurf.                                          *)
(*                                                                        *)
(*  GlSurf is free software; you can redistribute it and/or modify        *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation; either version 2 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  GlSurf is distributed in the hope that it will be useful,             *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with GlSurf; if not, write to the Free Software                 *)
(*  Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *)
(*  MA  02111-1307  USA                                                   *)
(**************************************************************************)

open Vect3
open Algebra
open Function
open Parameters

val all_surfaces : (string, surface) Hashtbl.t
val all_vectors : (string, point) Hashtbl.t ref

val do_delete : string -> unit
val do_begin : unit -> unit
val do_end : unit -> unit
val get_string : string -> string
val get_vector3 : point -> string -> point
val get_vector4 : point -> string -> point
val get_tuple3 : float * float * float -> string -> float * float * float
val get_tuple4 :
  float * float * float * float -> string -> float * float * float * float
val get_val : Expr.scalar -> string -> Expr.scalar
val get_bool : bool -> string -> bool
val fastest : float
val nicest : float
val dont_care : float
val off : float
val get_hint :
  Expr.scalar ->
  string -> [> `dont_care | `fastest | `nicest | `off ]
val get_int : int -> string -> int
val do_make_mesh : Expr.elem -> surface
(*
val do_make_mesh_correct : Expr.elem -> surface
*)
val do_make_curve :
  (compute -> vector * vector * vector -> unit) ->
  string option ->
  surface -> Expr.elem -> surface
val do_gl_start : unit -> unit
val do_draw : int -> surface list -> unit

type save_option =
    Binary | Ascii | Glsurf

val save_surface : save_option -> string -> surface -> unit
val load_surface : string -> surface
val create_point :
  string -> float array -> surface * [ `double ] Raw.t
val create_points :
  string -> float array list -> surface
val create_curve : string -> float array list -> surface
val move_point : [< Raw.fkind ] Raw.t -> float array -> unit
val all_surfaces : (string, surface) Hashtbl.t
val do_gl_start : unit -> unit

val adapt : surface -> Expr.elem -> float -> float -> float -> float -> surface
val do_adapt_curve :   surface -> Expr.elem -> surface

module Environment : Banach with
  type scalar = float and type elem = (string * float) list and
  type norm = float

val optimise : Expr.elem -> Expr.elem -> Expr.elem -> Expr.elem ->
  Environment.elem -> (float) Optimisation.parameters -> Environment.elem

val cur_stexture :    ([ `alpha | `blue | `green | `luminance | `red ] * [`chessboard | `lines] *
     (Vect3.vector -> float) * (Vect3.vector -> float) * (Vect3.vector -> float) * Gl.rgba *
     [ `blend | `decal | `modulate | `replace ])
    option ref

val cur_ctexture :    ([ `alpha | `blue | `green | `luminance | `red ] * [`chessboard | `lines] *
     (Vect3.vector -> float) * (Vect3.vector -> float) * (Vect3.vector -> float) * Gl.rgba *
     [ `blend | `decal | `modulate | `replace ])
    option ref

val newton_solve : Expr.elem list -> string list -> float list -> float array
