(**************************************************************************)
(*                                                                        *)
(*                               GlSurf                                   *)
(*                                                                        *)
(*                   C. Raffalli, Universite de Savoie                    *)
(*                                                                        *)
(* Copyright 2003, 2004 Christohe Raffalli                                *)
(*                                                                        *)
(*  This file is part of GlSurf.                                          *)
(*                                                                        *)
(*  GlSurf is free software; you can redistribute it and/or modify        *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation; either version 2 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  GlSurf is distributed in the hope that it will be useful,             *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *) 
(*  You should have received a copy of the GNU General Public License     *)
(*  along with GlSurf; if not, write to the Free Software                 *)
(*  Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *)
(*  MA  02111-1307  USA                                                   *)
(**************************************************************************)

open Format
open Vect3
open Function
open Parameters

    
type message = {
    message : string;
    message_x : int;
    message_y : int;
    message_color : float * float * float;
  }

val file_format : string ref
val save_image : string -> unit

type gl_action =
    Change_param of bool * gl_parameters * draw_parameters
  | Add_surfaces of surface list
  | Add_message of string * message
  | Remove_surfaces of surface list
  | Remove_message of string
  | Remove_all
  | Wait of int
  | LightOn of int
  | LightOff of int
  | LightPos of int * light_parameters
  | Redisplay
  | Curve_mode of bool
  | Save_image of string
  | Run of (unit -> unit)
  | Quit

val batch : bool ref
val gl_origin : point ref
val gl_size : float ref
val cube_edges : GlList.t option ref
val badt : triangle list ref
val gl_init : unit -> string array
val gl_started : Mutex.t
val draw : int -> gl_parameters -> draw_parameters -> surface list -> unit
val start_gl : gl_parameters -> draw_parameters -> unit
val queue_mutex : Mutex.t
val events_list : gl_action Queue.t
val add_to_queue : (gl_action list -> unit) ref
val string_params : (string * string ref) list
val float_params : (string * float ref) list
val int_params : (string * int ref) list
val bool_params : (string * bool ref) list
val time_before_next : int ref
val time_before_rotate : int ref
val curve_mode : bool ref
val init_pos : unit -> unit
val full_screen : bool -> unit
val lock_to_write : unit -> unit
