(**************************************************************************)
(*                                                                        *)
(*                               GlSurf                                   *)
(*                                                                        *)
(*                   C. Raffalli, Universite de Savoie                    *)
(*                                                                        *)
(* Copyright 2003, 2004 Christohe Raffalli                                *)
(*                                                                        *)
(*  This file is part of GlSurf.                                          *)
(*                                                                        *)
(*  GlSurf is free software; you can redistribute it and/or modify        *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation; either version 2 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  GlSurf is distributed in the hope that it will be useful,             *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with GlSurf; if not, write to the Free Software                 *)
(*  Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *)
(*  MA  02111-1307  USA                                                   *)
(**************************************************************************)

open Algebra
open Expression
open Vect3

val can_compile : bool
val cc_name : string
val cc_shopts : string

type handle
val reinit : handle * string -> unit
val dlclose : handle -> unit

module Expr : Expression with type scalar = float
  and type 'a environment = (string * 'a) list

module type Function =
  sig
    val value : point -> float
    val grad : point -> vector
    val hessian_mat : point -> float array
    val hessian : point -> vector -> vector
    val third : point -> vector -> float  array array
    val ihessian : float -> float -> float -> float -> float -> float ->
      (float * float) * (float * float) * (float * float) *
	(float * float) * (float * float) * (float * float)

    val handle : (handle * string) option
    val f : Expr.elem
    val fx : Expr.elem
    val fy : Expr.elem
    val fz : Expr.elem

    val epsilon : float
    val dighotomie : point -> point -> point
    val newton : float -> float -> point -> point
  end

module type String_Expr =
  sig
    val f : string
    val epsilon : float
    val compile : bool
    val hessian : bool
    val third : bool
 end

module type Expr_Expr =
  sig
    val f : Expr.elem
    val epsilon : float
    val compile : bool
    val hessian : bool
    val third : bool
  end

module Make_Function_From_String :
  functor (Str : String_Expr) -> Function

module Make_Function_From_Expr :
  functor (Str : Expr_Expr) -> Function
