(**************************************************************************)
(*                                                                        *)
(*                               GlSurf                                   *)
(*                                                                        *)
(*                   C. Raffalli, Universite de Savoie                    *)
(*                                                                        *)
(* Copyright 2003, 2004 Christohe Raffalli                                *)
(*                                                                        *)
(*  This file is part of GlSurf.                                          *)
(*                                                                        *)
(*  GlSurf is free software; you can redistribute it and/or modify        *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation; either version 2 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  GlSurf is distributed in the hope that it will be useful,             *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with GlSurf; if not, write to the Free Software                 *)
(*  Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *)
(*  MA  02111-1307  USA                                                   *)
(**************************************************************************)

type point = float array
type vector = point

(* the following two types are used only for colors and LablGL compatibility *)
type vector4 = float * float * float * float
type vector3 = float * float * float

val tuple : point -> float * float * float

type triangle = point * point * point

val ( ++ ) : vector -> vector -> vector

val ( -- ) : vector -> vector -> vector

val opp  : vector -> vector

val ( |. ) : vector -> vector -> float

(* product by a scalar *)
val ( ** ) : float -> vector -> vector

(* division by a scalar *)
val ( // ) : vector -> float -> vector

(* square of the euclidian norm *)
val norm2 : vector -> float

val norm : vector -> float

(* vectorial product *)
val vector_prod : vector -> vector -> vector

(* vectorial product *)
val project_vector : vector -> vector -> vector


(* discriminant of 3 vectors *)
val det : vector -> vector -> vector -> float

val tetra_vol : point -> point -> point -> point -> float

val middle : vector -> vector -> vector

val pi : float

(* angle of 2 vector (between -pi/2 and pi/2) *)
val angle : vector -> vector -> float

val print_vector : vector -> unit
val print_triangle : triangle -> unit

val in_cube : point -> point -> point -> bool

val in_cube_strict : point -> point -> point -> bool

(*
val merge_point : point -> point -> unit

val simplify_triangles : (point -> point -> bool) -> triangle list -> triangle list

val clean_triangles : triangle list -> triangle list
*)

val polygone_normal : point list -> vector

val triangles_from_polygones : triangle list -> point list list -> triangle list

val random_vector : unit -> vector

val random_normed_vector : unit -> vector

val hessian_mult : vector -> vector -> vector
